//+------------------------------------------------------------------+
//|                                                  allanmaug.mq5     |
//|                                  Copyright 2024, allanmaug         |
//|                                             https://www.allanmaug.com |
//+------------------------------------------------------------------+
#property copyright "Copyright 2024, allanmaug | Email: mwesigwaallanmaurice123@gmail.com"
#property link      "https://www.youtube.com/@all9nn"
#property version   "1.01"
#property strict

//+------------------------------------------------------------------+
//| Includes and Defines                                               |
//+------------------------------------------------------------------+
#include <Trade/Trade.mqh>
#include <Arrays/ArrayLong.mqh>
#define FILE_NAME MQLInfoString(MQL_PROGRAM_NAME)+".bin"

//+------------------------------------------------------------------+
//| Enums                                                              |
//+------------------------------------------------------------------+
enum ENUM_MODE {
    MODE_MASTER, // Master Mode
    MODE_SLAVE   // Slave Mode
};

//+------------------------------------------------------------------+
//| Expert Settings                                                    |
//+------------------------------------------------------------------+
string EAName = "COPIERX";                          // COPIER_NAME

input group "==== ACCOUNT MODE ===="
input ENUM_MODE Mode = MODE_SLAVE;                  // EA Mode

input group "==== SLAVE ACCOUNT Settings ===="
input bool   copySlandcopyTp = true;               // COPY_STOPLOSS_AND_TAKEPROFIT
input bool   takeoppositeTrades = false;           // TAKE_OPPOSITE_TRADES
input double lotMultiple = 1;                      // LOTSIZE_MULTIPLE
input int    MagicNumber = 12345;                  // MAGIC_NUMBER




ENUM_BASE_CORNER PanelCorner = CORNER_LEFT_UPPER;  // Panel Corner
int    PanelX = 20;                         // Panel X Position
int    PanelY = 100;                        // Panel Y Position
int    PanelWidth = 400;                    // Panel Width
int    PanelHeight = 150;                   // Panel Height

color  PanelColor = C'32,32,32';            // Panel Background Color
color  PanelBorder = clrGray;               // Panel Border Color
color  TextColor = clrWhite;                // Text Color
color  ProfitColor = clrLime;               // Profit Color
color  LossColor = clrRed;                  // Loss Color
int    FontSize = 10;                       // Font Size
string FontName = "Arial";                  // Font Name

int    TitleSpacing = 20;                   // Title Y Position
int    FirstRowY = 50;                      // First Row Y Position
int    RowSpacing = 20;                     // Space Between Rows

//+------------------------------------------------------------------+
//| Global Variables                                                   |
//+------------------------------------------------------------------+
datetime LastUpdateTime = 0;
CTrade trade;
CArrayLong copiedTickets;
bool tradeStatus[1];

//+------------------------------------------------------------------+
//| Expert Initialization Function                                     |
//+------------------------------------------------------------------+
int OnInit() {
    // Initialize trade magic number
    trade.SetExpertMagicNumber(MagicNumber);
    
    // Create appropriate dashboard based on mode
    if(Mode == MODE_SLAVE) {
        CreateSlaveDashboard();
    } else {
        CreateMasterDashboard();
    }
    
    // Initialize other components
    copiedTickets.Clear();
    EventSetMillisecondTimer(50);
    return(INIT_SUCCEEDED);
}

//+------------------------------------------------------------------+
//| Expert Deinitialization Function                                   |
//+------------------------------------------------------------------+
void OnDeinit(const int reason) {
    EventKillTimer();
    copiedTickets.Clear();
    Comment("");
    ObjectsDeleteAll(0, "Dashboard_");
    ObjectsDeleteAll(0, "MasterDash_");
    DeleteEAFile();
}


void OnTimer(){
    if(Mode == MODE_MASTER) {
        UpdateMasterDashboard();
        int file = FileOpen(FILE_NAME, FILE_WRITE|FILE_BIN|FILE_COMMON);
        if(file != INVALID_HANDLE){
            for(int i = 0; i < PositionsTotal(); i++) {
                ulong ticket = PositionGetTicket(i);
                if(PositionSelectByTicket(ticket)) {
                    string symbol = PositionGetString(POSITION_SYMBOL);

                        FileWriteLong(file, ticket);
                        int length = StringLen(symbol);
                        FileWriteInteger(file, length);
                        FileWriteString(file, symbol);
                        FileWriteDouble(file, PositionGetDouble(POSITION_VOLUME));
                        FileWriteInteger(file, (int)PositionGetInteger(POSITION_TYPE));
                        FileWriteDouble(file, PositionGetDouble(POSITION_PRICE_OPEN));
                        FileWriteDouble(file, PositionGetDouble(POSITION_SL));
                        FileWriteDouble(file, PositionGetDouble(POSITION_TP));
                    
                }
            }
            FileClose(file);
        }
    }
    else if(Mode == MODE_SLAVE){
        CArrayLong currentMasterTickets;
        currentMasterTickets.Clear();
        UpdateSlaveDashboard();
        int file = FileOpen(FILE_NAME, FILE_READ|FILE_BIN|FILE_COMMON);
        if(file != INVALID_HANDLE){
            while(!FileIsEnding(file)) {
                long masterTicket = FileReadLong(file);
                int length = FileReadInteger(file);
                string masterSymbol = FileReadString(file, length);
                double volume = FileReadDouble(file);
                ENUM_POSITION_TYPE posType = (ENUM_POSITION_TYPE)FileReadInteger(file);
                double openPrice = FileReadDouble(file);
                double sl = (copySlandcopyTp == true) ? FileReadDouble(file) : 0; 
                double tp = (copySlandcopyTp == true) ? FileReadDouble(file) : 0;
                
                string slaveSymbol = FindMatchingSymbol(masterSymbol);
                if(slaveSymbol == "") continue;  
                
                currentMasterTickets.Add(masterTicket);
                
                if((copiedTickets.SearchLinear(masterTicket) == -1) && (CheckMasterComment(masterTicket) == false)) {
                    Print("New position detected: ", masterSymbol, " -> ", slaveSymbol);
                    if(posType == POSITION_TYPE_BUY){
                        if(takeoppositeTrades == false){
                           if(trade.Buy(ValidateLotSize(volume*lotMultiple), slaveSymbol, 0, sl, tp, IntegerToString(masterTicket))){
                               copiedTickets.Add(masterTicket);
                               Print("Opened Buy position for ticket: ", masterTicket, " on ", slaveSymbol);
                           }
                        }
                        if(takeoppositeTrades == true){
                           if(trade.Sell(ValidateLotSize(volume*lotMultiple), slaveSymbol, 0, tp, sl, IntegerToString(masterTicket))){
                               copiedTickets.Add(masterTicket);
                               Print("Opened Buy position for ticket: ", masterTicket, " on ", slaveSymbol);
                           }
                        }
                    }
                    else if(posType == POSITION_TYPE_SELL) {
                        if(takeoppositeTrades == false){
                           if(trade.Sell(ValidateLotSize(volume*lotMultiple), slaveSymbol, 0, sl, tp, IntegerToString(masterTicket))) {
                               copiedTickets.Add(masterTicket);
                               Print("Opened Sell position for ticket: ", masterTicket, " on ", slaveSymbol);
                           }
                        }
                        if(takeoppositeTrades == true){
                           if(trade.Buy(ValidateLotSize(volume*lotMultiple), slaveSymbol, 0, tp, sl, IntegerToString(masterTicket))) {
                               copiedTickets.Add(masterTicket);
                               Print("Opened Sell position for ticket: ", masterTicket, " on ", slaveSymbol);
                           }
                        }
                    }
                }
                else {
                    for(int i = 0; i < PositionsTotal(); i++) {
                        ulong ticket = PositionGetTicket(i);
                        if(PositionSelectByTicket(ticket)) {
                            if(PositionGetString(POSITION_SYMBOL) == slaveSymbol && 
                               PositionGetInteger(POSITION_MAGIC) == MagicNumber && 
                               StringToInteger(PositionGetString(POSITION_COMMENT)) == masterTicket){
                                if(takeoppositeTrades == false){             
                                   if(PositionGetDouble(POSITION_SL) != sl || PositionGetDouble(POSITION_TP) != tp){
                                       trade.PositionModify(ticket, sl, tp);
                                   }
                                }
                                if(takeoppositeTrades == true){             
                                   if(PositionGetDouble(POSITION_SL) != tp || PositionGetDouble(POSITION_TP) != sl){
                                       trade.PositionModify(ticket, tp, sl);
                                   }
                                }
                                break;
                            }
                        }
                    }
                }
            }
            FileClose(file);
            
            for(int i = 0; i < PositionsTotal(); i++){
                ulong ticket = PositionGetTicket(i);
                if(PositionSelectByTicket(ticket)){
                    string symbol = PositionGetString(POSITION_SYMBOL);
                    if(PositionGetInteger(POSITION_MAGIC) == MagicNumber){  
                        long masterTicket = StringToInteger(PositionGetString(POSITION_COMMENT));
                        if(currentMasterTickets.SearchLinear(masterTicket) == -1) {
                            if(trade.PositionClose(ticket)) {
                                copiedTickets.Delete(copiedTickets.SearchLinear(masterTicket));
                                Print("Closed position for ticket: ", masterTicket, " on ", symbol);
                            }
                        }
                    }
                }
            }
        }
    }
}

// Utility Functions
//+------------------------------------------------------------------+
double ValidateLotSize(double lots){
    double minLot = SymbolInfoDouble(_Symbol, SYMBOL_VOLUME_MIN);
    double maxLot = SymbolInfoDouble(_Symbol, SYMBOL_VOLUME_MAX);
    double stepLot = SymbolInfoDouble(_Symbol, SYMBOL_VOLUME_STEP);
    lots = MathMax(MathMin(lots, maxLot), minLot);
    int digits = (int)MathLog10(1.0 / stepLot);
    return NormalizeDouble(lots, digits);
}

bool CheckMasterComment(long master_comment){
   int total_positions = PositionsTotal();   
   if(total_positions == 0)
      return false;
   for(int i = 0; i < total_positions; i++){
      if(PositionSelect(PositionGetSymbol(i))){
         if(PositionGetString(POSITION_COMMENT) == (string)master_comment)
            return true;
            break;
      }
   }
   return false;
}

int GetCopiedTradeCount() {
    return copiedTickets.Total();
}

bool DeleteEAFile() {
    string filename = MQLInfoString(MQL_PROGRAM_NAME) + ".bin";
    if(FileIsExist(filename, FILE_COMMON)) {
        bool deleteResult = FileDelete(filename, FILE_COMMON);
        if(deleteResult){
            Print("Successfully deleted file: ", filename);
            return true;
        } else {
            Print("Failed to delete file: ", filename, ". Error code: ", GetLastError());
            return false;
        }
    }
    Print("File does not exist: ", filename);
    return false;
}

bool Doesmasterfileexist(){
    string filename = MQLInfoString(MQL_PROGRAM_NAME) + ".bin";
    if(FileIsExist(filename, FILE_COMMON)) {
      return true;
    }
    return false;
}

//+------------------------------------------------------------------+
//| Slave Dashboard Functions                                          |
//+------------------------------------------------------------------+
void CreateSlaveDashboard() {
    ObjectsDeleteAll(0, "Dashboard_");
    CreateSlavePanel();
    
    long accountNumber = AccountInfoInteger(ACCOUNT_LOGIN);
    string title = EAName + " - " + IntegerToString(accountNumber) + " - RECEIVER";
    
    // Create labels with proper spacing
    CreateSlaveLabel("Title", PanelX + 10, PanelY + TitleSpacing, title, clrGold, 12);
    
    // Status Information
    CreateSlaveLabel("ConnectionStatus", PanelX + 10, PanelY + FirstRowY, "Connection Status: ", TextColor);
    CreateSlaveLabel("TradeCount", PanelX + 10, PanelY + FirstRowY + RowSpacing, "Total Trades: ", TextColor);
    CreateSlaveLabel("LastTrade", PanelX + 10, PanelY + FirstRowY + (RowSpacing * 2), "Last Trade Status: ", TextColor);
    
    // Settings Information
    CreateSlaveLabel("CopySL", PanelX + 10, PanelY + FirstRowY + (RowSpacing * 3), "Copy SL/TP: ", TextColor);
    CreateSlaveLabel("OppositeMode", PanelX + 10, PanelY + FirstRowY + (RowSpacing * 4), "Opposite Trades: ", TextColor);
    CreateSlaveLabel("LotMultiple", PanelX + 10, PanelY + FirstRowY + (RowSpacing * 5), "Lot Multiplier: ", TextColor);
    
    // Create value labels
    CreateSlaveLabel("ConnectionValue", PanelX + 150, PanelY + FirstRowY, "", TextColor);
    CreateSlaveLabel("TradeCountValue", PanelX + 150, PanelY + FirstRowY + RowSpacing, "", TextColor);
    CreateSlaveLabel("LastTradeValue", PanelX + 150, PanelY + FirstRowY + (RowSpacing * 2), "", TextColor);
    CreateSlaveLabel("CopySLValue", PanelX + 150, PanelY + FirstRowY + (RowSpacing * 3), "", TextColor);
    CreateSlaveLabel("OppositeModeValue", PanelX + 150, PanelY + FirstRowY + (RowSpacing * 4), "", TextColor);
    CreateSlaveLabel("LotMultipleValue", PanelX + 150, PanelY + FirstRowY + (RowSpacing * 5), "", TextColor);
    
    ObjectSetInteger(0, "Dashboard_BG", OBJPROP_YSIZE, PanelHeight + (RowSpacing * 3));
}

void CreateSlavePanel() {
    ObjectCreate(0, "Dashboard_BG", OBJ_RECTANGLE_LABEL, 0, 0, 0);
    ObjectSetInteger(0, "Dashboard_BG", OBJPROP_XDISTANCE, PanelX);
    ObjectSetInteger(0, "Dashboard_BG", OBJPROP_YDISTANCE, PanelY);
    ObjectSetInteger(0, "Dashboard_BG", OBJPROP_XSIZE, PanelWidth);
    ObjectSetInteger(0, "Dashboard_BG", OBJPROP_YSIZE, PanelHeight);
    ObjectSetInteger(0, "Dashboard_BG", OBJPROP_BGCOLOR, PanelColor);
    ObjectSetInteger(0, "Dashboard_BG", OBJPROP_BORDER_TYPE, BORDER_FLAT);
    ObjectSetInteger(0, "Dashboard_BG", OBJPROP_CORNER, PanelCorner);
    ObjectSetInteger(0, "Dashboard_BG", OBJPROP_BORDER_COLOR, PanelBorder);
    ObjectSetInteger(0, "Dashboard_BG", OBJPROP_SELECTABLE, false);
    ObjectSetInteger(0, "Dashboard_BG", OBJPROP_SELECTED, false);
    ObjectSetInteger(0, "Dashboard_BG", OBJPROP_HIDDEN, true);
}

void CreateSlaveLabel(string name, int x, int y, string text, color clr=CLR_NONE, int size=-1) {
    if(clr == CLR_NONE) clr = TextColor;
    if(size == -1) size = FontSize;
    
    ObjectCreate(0, "Dashboard_" + name, OBJ_LABEL, 0, 0, 0);
    ObjectSetString(0, "Dashboard_" + name, OBJPROP_TEXT, text);
    ObjectSetInteger(0, "Dashboard_" + name, OBJPROP_XDISTANCE, x);
    ObjectSetInteger(0, "Dashboard_" + name, OBJPROP_YDISTANCE, y);
    ObjectSetInteger(0, "Dashboard_" + name, OBJPROP_CORNER, PanelCorner);
    ObjectSetInteger(0, "Dashboard_" + name, OBJPROP_COLOR, clr);
    ObjectSetString(0, "Dashboard_" + name, OBJPROP_FONT, FontName);
    ObjectSetInteger(0, "Dashboard_" + name, OBJPROP_FONTSIZE, size);
    ObjectSetInteger(0, "Dashboard_" + name, OBJPROP_SELECTABLE, false);
    ObjectSetInteger(0, "Dashboard_" + name, OBJPROP_SELECTED, false);
    ObjectSetInteger(0, "Dashboard_" + name, OBJPROP_HIDDEN, true);
}

void UpdateSlaveLabel(string name, string text, color clr=CLR_NONE) {
    if(clr == CLR_NONE) clr = TextColor;
    if(ObjectFind(0, "Dashboard_" + name) >= 0) {
        ObjectSetString(0, "Dashboard_" + name, OBJPROP_TEXT, text);
        ObjectSetInteger(0, "Dashboard_" + name, OBJPROP_COLOR, clr);
    }
}

void UpdateSlaveDashboard() {
    if(TimeLocal() - LastUpdateTime < 1) return;
    LastUpdateTime = TimeLocal();
    
    // Update connection status
    string connectionStatus = Doesmasterfileexist() && IsAlgoTradingAllowed() ? "CONNECTED" : "DISCONNECTED";
    color connectionColor = Doesmasterfileexist() ? ProfitColor : LossColor;
    UpdateSlaveLabel("ConnectionValue", connectionStatus, connectionColor);
    
    // Update trade count and status
    int tradeCount = GetCopiedTradeCount();
    UpdateSlaveLabel("TradeCountValue", IntegerToString(tradeCount));
    
    string lastTradeStatus = tradeCount > 0 ? "Active" : "No Trades";
    color statusColor = tradeCount > 0 ? ProfitColor : TextColor;
    UpdateSlaveLabel("LastTradeValue", lastTradeStatus, statusColor);
    
    // Update settings status
    UpdateSlaveLabel("CopySLValue", copySlandcopyTp ? "ENABLED" : "DISABLED", copySlandcopyTp ? ProfitColor : LossColor);
    UpdateSlaveLabel("OppositeModeValue", takeoppositeTrades ? "ENABLED" : "DISABLED", takeoppositeTrades ? ProfitColor : TextColor);
    UpdateSlaveLabel("LotMultipleValue", DoubleToString(lotMultiple, 2) + "x", lotMultiple == 1.0 ? TextColor : ProfitColor);
    
    ChartRedraw();
}

//+------------------------------------------------------------------+
//| Master Dashboard Functions                                         |
//+------------------------------------------------------------------+
void CreateMasterDashboard() {
    ObjectsDeleteAll(0, "MasterDash_");
    CreateMasterPanel();
    
    long accountNumber = AccountInfoInteger(ACCOUNT_LOGIN);
    string title = EAName + " - " + IntegerToString(accountNumber) + " - TRANSMITTER";
    
    CreateMasterLabel("Title", PanelX + 10, PanelY + TitleSpacing, title, clrGold, 12);
    CreateMasterLabel("TradesLabel", PanelX + 10, PanelY + FirstRowY, "Trades Sent: ", TextColor);
    CreateMasterLabel("TradesValue", PanelX + 150, PanelY + FirstRowY, "", TextColor);
    
    ObjectSetInteger(0, "MasterDash_BG", OBJPROP_YSIZE, 100);
}

void CreateMasterPanel() {
    ObjectCreate(0, "MasterDash_BG", OBJ_RECTANGLE_LABEL, 0, 0, 0);
    ObjectSetInteger(0, "MasterDash_BG", OBJPROP_XDISTANCE, PanelX);
    ObjectSetInteger(0, "MasterDash_BG", OBJPROP_YDISTANCE, PanelY);
    ObjectSetInteger(0, "MasterDash_BG", OBJPROP_XSIZE, PanelWidth);
    ObjectSetInteger(0, "MasterDash_BG", OBJPROP_YSIZE, PanelHeight);
    ObjectSetInteger(0, "MasterDash_BG", OBJPROP_BGCOLOR, PanelColor);
    ObjectSetInteger(0, "MasterDash_BG", OBJPROP_BORDER_TYPE, BORDER_FLAT);
    ObjectSetInteger(0, "MasterDash_BG", OBJPROP_CORNER, PanelCorner);
    ObjectSetInteger(0, "MasterDash_BG", OBJPROP_BORDER_COLOR, PanelBorder);
    ObjectSetInteger(0, "MasterDash_BG", OBJPROP_SELECTABLE, false);
    ObjectSetInteger(0, "MasterDash_BG", OBJPROP_SELECTED, false);
    ObjectSetInteger(0, "MasterDash_BG", OBJPROP_HIDDEN, true);
}

void CreateMasterLabel(string name, int x, int y, string text, color clr=CLR_NONE, int size=-1) {
    if(clr == CLR_NONE) clr = TextColor;
    if(size == -1) size = FontSize;
    
    ObjectCreate(0, "MasterDash_" + name, OBJ_LABEL, 0, 0, 0);
    ObjectSetString(0, "MasterDash_" + name, OBJPROP_TEXT, text);
    ObjectSetInteger(0, "MasterDash_" + name, OBJPROP_XDISTANCE, x);
    ObjectSetInteger(0, "MasterDash_" + name, OBJPROP_YDISTANCE, y);
    ObjectSetInteger(0, "MasterDash_" + name, OBJPROP_CORNER, PanelCorner);
    ObjectSetInteger(0, "MasterDash_" + name, OBJPROP_COLOR, clr);
    ObjectSetString(0, "MasterDash_" + name, OBJPROP_FONT, FontName);
    ObjectSetInteger(0, "MasterDash_" + name, OBJPROP_FONTSIZE, size);
    ObjectSetInteger(0, "MasterDash_" + name, OBJPROP_SELECTABLE, false);
    ObjectSetInteger(0, "MasterDash_" + name, OBJPROP_SELECTED, false);
    ObjectSetInteger(0, "MasterDash_" + name, OBJPROP_HIDDEN, true);
}

void UpdateMasterLabel(string name, string text, color clr=CLR_NONE) {
    if(clr == CLR_NONE) clr = TextColor;
    if(ObjectFind(0, "MasterDash_" + name) >= 0) {
        ObjectSetString(0, "MasterDash_" + name, OBJPROP_TEXT, text);
        ObjectSetInteger(0, "MasterDash_" + name, OBJPROP_COLOR, clr);
    }
}

void UpdateMasterDashboard() {
    if(TimeLocal() - LastUpdateTime < 1) return;
    LastUpdateTime = TimeLocal();
    
    UpdateMasterLabel("TradesValue", IntegerToString(PositionsTotal()));
    ChartRedraw();
}

//+------------------------------------------------------------------+
//| Check if algorithmic trading is allowed                           |
//+------------------------------------------------------------------+
bool IsAlgoTradingAllowed() {
   if(!TerminalInfoInteger(TERMINAL_TRADE_ALLOWED)) {
      PrintFormat("Error: Automated trading is disabled in terminal");
      return false;
   }
   
   if(!MQLInfoInteger(MQL_TRADE_ALLOWED)) {
      PrintFormat("Error: Algorithmic trading is disabled by MetaTrader settings");
      return false;
   }
   
   if(!AccountInfoInteger(ACCOUNT_TRADE_EXPERT)) {
      PrintFormat("Error: Automated trading is disabled for this account");
      return false;
   }
   
   if(!AccountInfoInteger(ACCOUNT_TRADE_ALLOWED)) {
      PrintFormat("Error: Trading is disabled for this account");
      return false;
   }
   
   return true;
}

string FindMatchingSymbol(string searchSymbol){
    string bestMatch = "";
    int totalSymbols = SymbolsTotal(false);
    double bestSimilarity = 0.0;
    // Create uppercase version of search symbol for comparison
    string upperSearchSymbol = searchSymbol;
    StringToUpper(upperSearchSymbol);
    for(int i = 0; i < totalSymbols; i++){
        string originalSymbol = SymbolName(i, false);
        string upperSymbol = originalSymbol;
        StringToUpper(upperSymbol);
        if(upperSymbol == upperSearchSymbol){
            return originalSymbol;
        }
        double similarity = 0.0;
        int maxLen = MathMax(StringLen(upperSymbol), StringLen(upperSearchSymbol));
        int minLen = MathMin(StringLen(upperSymbol), StringLen(upperSearchSymbol));
        for(int j = 0; j < minLen; j++){
            if(StringGetCharacter(upperSymbol, j) == StringGetCharacter(upperSearchSymbol, j)){
                similarity += 1.0;
            }
        }
        if(maxLen > 0){
            similarity = similarity / maxLen;
        }
        if(similarity > bestSimilarity){
            bestSimilarity = similarity;
            bestMatch = originalSymbol;
        }
    } 
    return (bestSimilarity > 0.5) ? bestMatch : "";
}